

#################################################
# Additional constructors
#################################################

# A constructor for normal toric varieties of picard number one.
# The fan will be the unique complete fan having `transpose(P)` as its generator matrix.
function normal_toric_variety(P :: MatElem; non_redundant :: Bool)
    n, r = nrows(P), ncols(P)
    @req n == r + 1 "this constructor only works if the picard rank is one"
    cones = collect(powerset(1 : n, r, r))
    normal_toric_variety(P, cones, non_redundant=non_redundant)
end

@attr function affine_toric_charts(X :: Oscar.NormalToricVarietyType)
    P = gen_matrix(X)
    charts = Dict{Vector{Int}, NormalToricVariety}()
    for c in maximal_cones_indices(X)
        charts[c] = normal_toric_variety(transpose(P[:, c]), [collect(1:length(c))]; non_redundant=true)
    end
    return charts
end


# return the affine toric chart associated with the given cone
affine_toric_chart(X :: Oscar.NormalToricVarietyType, c :: Vector{Int64}) = affine_toric_charts(X)[c]

# remove the torus factor of a toric variety.
# This means using the rational vector space generated by the rays of the fan
# as its ambient space.
function remove_torusfactor(X :: Oscar.NormalToricVarietyType)
    P = gen_matrix(X)
    r, A, _ = rcef_rational(P)
    A = A[:, 1 : r]
    # Use `invoke` here to dispatch on the more generic `solve_rational` from 
    # AbstractAlgebra, which doesn't require the first argument to be a square matrix
    new_P, _ = invoke(solve_rational, Tuple{MatElem{ZZRingElem}, MatElem{ZZRingElem}}, A, P)
    normal_toric_variety(transpose(new_P), maximal_cones_indices(X); non_redundant=true)
end

#################################################
# Local class groups
#################################################

@attr local_class_groups(X :: Oscar.NormalToricVarietyType) =
Dict([(c, class_group(U)) for (c,U) in affine_toric_charts(X)])

local_class_group(X :: Oscar.NormalToricVarietyType, c :: Vector{Int64}) =
local_class_groups(X)[c]

@attr function maps_from_class_group_to_local_class_groups(X :: Oscar.NormalToricVarietyType)
    K, f = cokernel(map_from_character_lattice_to_torusinvariant_weil_divisor_group(X))
    maps = Dict{Vector{Int}, GrpAbFinGenMap}()
    for c in maximal_cones_indices(X)
        U = affine_toric_chart(X, c)
        KU, fU = cokernel(map_from_character_lattice_to_torusinvariant_weil_divisor_group(U))
        m = zero_matrix(ZZ, nrays(X), nrays(U))
        for i in 1:length(c)
            m[c[i],i] = 1
        end
        f = snf(K)[2] * hom(K, KU, m) * inv(snf(KU)[2])
        maps[c] = hom(class_group(X), local_class_group(X,c), matrix(f))
    end
    return maps
end

map_from_class_group_to_local_class_group(X :: Oscar.NormalToricVarietyType, c :: Vector{Int64}) =
maps_from_class_group_to_local_class_groups(X)[c]


#################################################
# Cones in the rational class group
#################################################


@attr effective_cone(X :: Oscar.NormalToricVarietyType) =
positive_hull(transpose(degree_matrix_free_part(X)))

@attr function moving_cone(X :: Oscar.NormalToricVarietyType)
    Q0 = degree_matrix_free_part(X)
    n = nrays(X)
    intersect([positive_hull(transpose(Q0[:,c])) for c in powerset(1:n, n-1, n-1)])
end

@attr function semiample_cone(X :: Oscar.NormalToricVarietyType)
    Q0 = degree_matrix_free_part(X)
    intersect([positive_hull(transpose(Q0[:,c])) for c in covering_collection(X)])
end

    

